"""
Drive and manufacturer-specific quirk database.

This module contains a database of drive and manufacturer-specific quirks,
primarily used for vendor-specific SMART data. This is necessary because
SMART is a specification for _communicating_ with a drive, not a specification
of what SMART data actually means.

.. note::

    It may be very tempting to copy data from sources like smartmontools, which
    has a database of drive-specific quirks. However, these databases are often
    licensed under the GPL, which is incompatible with the MIT license used by
    SMARTie. As such, this database is built from vendor spec sheets when
    possible and guesswork when it's not. If you wish to contribute to this
    database, please do so by submitting a pull request on GitHub.

"""

import enum
import re
from dataclasses import field, dataclass
from typing import Dict, List, Optional


class Units(enum.IntEnum):
    """
    Hints for the possible measurement unit of a SMART attribute, primarily
    used for display purposes.
    """

    UNKNOWN = 0
    CELSIUS = 10
    MILLISECONDS = 20
    HOURS = 21
    COUNT = 30


@dataclass
class SMARTAttribute:
    """
    An Attribute represents a single parsed SMART attribute.
    """

    #: A human-readable identifier, if known.
    name: str
    #: The SMART Attribute ID.
    id: int
    #: The SMART Attribute flags.
    flags: int = 0
    #: If known, a hint for the possible measurement unit in `current_value`.
    unit: Units = Units.UNKNOWN
    #: If provided, a callable function which will be used to process
    #: `current_value` and `worst_value`.
    processor: Optional[callable] = None
    #: The current value of the attribute.
    current_value: Optional[int] = None
    #: The worst known value of the attribute.
    worst_value: Optional[int] = None
    #: The maximum acceptable value of the attribute.
    threshold: Optional[int] = None

    @property
    def p_value(self):
        """
        The current value, run through any provided processor.
        """
        if self.processor is not None:
            return self.processor(self.current_value)
        return self.current_value

    @property
    def p_worst_value(self):
        """
        The worst value, run through any provided processor.
        """
        if self.processor is not None:
            return self.processor(self.worst_value)
        return self.worst_value


@dataclass
class DriveEntry:
    """
    Represents a single entry in the drive database.

    This is used to provide non-standard SMART attributes for specific drives,
    and other drive-specific or manufacturer-specific information.
    """

    name: str
    # A list of filters that a drive must match to use this entry.
    filters: List[str] = field(default_factory=lambda: ["*"])
    # A list of smart attributes that are specific to this drive.
    smart_attributes: Dict[int, SMARTAttribute] = field(default_factory=dict)
    # Maintainer notes.
    notes: List[str] = field(default_factory=list)


DRIVE_DATABASE = [
    DriveEntry(
        name="Default",
        filters=["type:ata"],
        smart_attributes={
            0x01: SMARTAttribute("READ_ERROR_RATE", 0x01),
            0x02: SMARTAttribute("THROUGHPUT_PERFORMANCE", 0x02),
            0x03: SMARTAttribute("SPIN_UP_TIME", 0x03, unit=Units.MILLISECONDS),
            0x04: SMARTAttribute("START_STOP_COUNT", 0x04, unit=Units.COUNT),
            0x05: SMARTAttribute(
                "REALLOCATED_SECTORS_COUNT", 0x05, unit=Units.COUNT
            ),
            0x06: SMARTAttribute("READ_CHANNEL_MARGIN", 0x06),
            0x07: SMARTAttribute("SEEK_ERROR_RATE", 0x07),
            0x08: SMARTAttribute("SEEK_TIME_PERFORMANCE", 0x08),
            # This name is a lie, some manufactures may store minutes or seconds
            # instead of hours.
            0x09: SMARTAttribute("POWER_ON_HOURS", 0x09, unit=Units.HOURS),
            0x0A: SMARTAttribute("SPIN_RETRY_COUNT", 0x0A, unit=Units.COUNT),
            0x0B: SMARTAttribute(
                "RECALIBRATION_RETRIES", 0x0B, unit=Units.COUNT
            ),
            0x0C: SMARTAttribute("POWER_CYCLE_COUNT", 0x0C, unit=Units.COUNT),
            0x0D: SMARTAttribute(
                "SOFT_READ_ERROR_RATE", 0x0D, unit=Units.COUNT
            ),
            0x16: SMARTAttribute("CURRENT_HELIUM_LEVEL", 0x16),
            # TODO: See E8
            0xAA: SMARTAttribute("AVAILABLE_RESERVED_SPACE", 0xAA),
            0xAB: SMARTAttribute(
                "SSD_PROGRAM_FAIL_COUNT", 0xAB, unit=Units.COUNT
            ),
            0xAC: SMARTAttribute(
                "SSD_ERASE_FAIL_COUNT", 0xAC, unit=Units.COUNT
            ),
            0xAD: SMARTAttribute(
                "SSD_WEAR_LEVELING_COUNT", 0xAD, unit=Units.COUNT
            ),
            0xAE: SMARTAttribute("POWER_LOSS_COUNT", 0xAE, unit=Units.COUNT),
            # 0xAF: Attribute('POWER_LOSS_PROTECTION_FAILURE', 0xAF),
            0xB0: SMARTAttribute("ERASE_FAIL_COUNT", 0xB0, unit=Units.COUNT),
            0xB1: SMARTAttribute("WEAR_RANGE_DELTA", 0xB1),
            0xB2: SMARTAttribute(
                "USED_RESERVED_BLOCK_COUNT", 0xB2, unit=Units.COUNT
            ),
            0xB3: SMARTAttribute(
                "USED_RESERVED_BLOCK_COUNT_TOTAL", 0xB3, unit=Units.COUNT
            ),
            0xB4: SMARTAttribute(
                "UNUSED_RESERVED_BLOCK_COUNT_TOTAL", 0xB4, unit=Units.COUNT
            ),
            0xB5: SMARTAttribute(
                "PROGRAM_FAIL_COUNT_TOTAL", 0xB5, unit=Units.COUNT
            ),
            0xB6: SMARTAttribute("ERASE_FAIL_COUNT", 0xB6, unit=Units.COUNT),
            0xB7: SMARTAttribute("RUNTIME_BAD_BLOCK", 0xB7, unit=Units.COUNT),
            0xB8: SMARTAttribute("PARITY_ERROR_COUNT", 0xB8, unit=Units.COUNT),
            0xB9: SMARTAttribute("HEAD_STABILITY", 0xB9),
            0xBA: SMARTAttribute("INDUCED_OP_VIBRATION_DETECTION", 0xBA),
            0xBB: SMARTAttribute(
                "REPORTED_UNCORRECTABLE_ERRORS", 0xBB, unit=Units.COUNT
            ),
            0xBC: SMARTAttribute("COMMANDS_TIMED_OUT", 0xBC, unit=Units.COUNT),
            0xBD: SMARTAttribute("HIGH_FLY_WRITES", 0xBD, unit=Units.COUNT),
            0xBE: SMARTAttribute(
                "TEMPERATURE_DIFFERENCE",
                0xBE,
                unit=Units.CELSIUS,
                processor=lambda v: 100 - v,
            ),
            0xBF: SMARTAttribute("GSENSE_ERROR_RATE", 0xBF, unit=Units.COUNT),
            0xC0: SMARTAttribute(
                "UNSAFE_SHUTDOWN_COUNT", 0xC0, unit=Units.COUNT
            ),
            0xC1: SMARTAttribute("LOAD_CYCLE_COUNT", 0xC1, unit=Units.COUNT),
            0xC2: SMARTAttribute(
                "TEMPERATURE_ABSOLUTE", 0xC2, unit=Units.CELSIUS
            ),
            0xC3: SMARTAttribute("HARDWARE_ECC_RECOVERED", 0xC3),
            0xC4: SMARTAttribute(
                "REALLOCATION_EVENT_COUNT", 0xC4, unit=Units.COUNT
            ),
            0xC5: SMARTAttribute(
                "CURRENT_PENDING_SECTOR_COUNT", 0xC5, unit=Units.COUNT
            ),
            0xC6: SMARTAttribute(
                "UNCORRECTABLE_SECTOR_COUNT", 0xC6, unit=Units.COUNT
            ),
            0xC7: SMARTAttribute(
                "ULTRA_DMA_CRC_ERROR_COUNT", 0xC7, unit=Units.COUNT
            ),
            0xC8: SMARTAttribute("WRITE_ERROR_RATE", 0xC8, unit=Units.COUNT),
            0xC9: SMARTAttribute(
                "SOFT_READ_ERROR_RATE", 0xC9, unit=Units.COUNT
            ),
            0xCA: SMARTAttribute("DATA_ADDRESS_MARKS", 0xCA, unit=Units.COUNT),
            0xCB: SMARTAttribute("RUN_OUT_CANCEL", 0xCB, unit=Units.COUNT),
            0xCC: SMARTAttribute("SOFT_ECC_CORRECTION", 0xCC, unit=Units.COUNT),
            0xCD: SMARTAttribute(
                "THERMAL_ASPERITY_RATE", 0xCD, unit=Units.COUNT
            ),
            0xCE: SMARTAttribute("FLYING_HEIGHT", 0xCE),
            0xCF: SMARTAttribute("SPIN_HEIGHT_CURRENT", 0xCF),
            0xD0: SMARTAttribute("SPIN_BUZZ", 0xD0, unit=Units.COUNT),
            0xD1: SMARTAttribute("OFFLINE_SEEK_PERFORMANCE", 0xD1),
            0xD2: SMARTAttribute("VIBRATION_DURING_WRITE", 0xD2),
            0xD3: SMARTAttribute("VIBRATION_DURING_WRITE", 0xD3),
            0xD4: SMARTAttribute("SHOCK_DURING_WRITE", 0xD4),
            0xDC: SMARTAttribute("DISK_SHIFT", 0xDC),
            0xDD: SMARTAttribute("GSENSE_ERROR_RATE", 0xDD, unit=Units.COUNT),
            0xDE: SMARTAttribute("LOADED_HOURS", 0xDE, unit=Units.HOURS),
            0xDF: SMARTAttribute(
                "LOAD_UNLOAD_RETRY_COUNT", 0xDF, unit=Units.COUNT
            ),
            0xE0: SMARTAttribute("LOAD_FRICTION", 0xE0),
            0xE1: SMARTAttribute(
                "LOAD_UNLOAD_CYCLE_COUNT", 0xE1, unit=Units.COUNT
            ),
            0xE2: SMARTAttribute("LOAD_IN_TIME", 0xE2),
            0xE3: SMARTAttribute(
                "TORQUE_AMPLIFICATION_COUNT", 0xE3, unit=Units.COUNT
            ),
            0xE4: SMARTAttribute(
                "POWER_OFF_RETRACT_CYCLE", 0xE4, unit=Units.COUNT
            ),
            0xE6: SMARTAttribute("THRASHING", 0xE6),
            0xE7: SMARTAttribute("LIFE_LEFT", 0xE7),
            0xE8: SMARTAttribute("ENDURANCE_REMAINING", 0xE8),
            0xE9: SMARTAttribute("MEDIA_WEAROUT_INDICATOR", 0xE9),
            # 0xEA
            # 0xEB
            0xF0: SMARTAttribute("HEAD_FLYING_HOURS", 0xF0, unit=Units.HOURS),
            0xF1: SMARTAttribute("TOTAL_LBAS_WRITTEN", 0xF1, unit=Units.COUNT),
            0xF2: SMARTAttribute("TOTAL_LBAS_READ", 0xF2, unit=Units.COUNT),
            0xF3: SMARTAttribute("TOTAL_LBAS_WRITTEN_EX", 0xF3),
            0xF4: SMARTAttribute("TOTAL_LBAS_READ_EX", 0xF4),
            0xF9: SMARTAttribute("NAND_WRITES", 0xF9),
            0xFA: SMARTAttribute(
                "READ_ERROR_RETRY_RATE", 0xFA, unit=Units.COUNT
            ),
            0xFB: SMARTAttribute("MINIMUM_SPARES_REMAINING", 0xFB),
            0xFC: SMARTAttribute("NEWLY_ADDED_BAD_FLASH_BLOCk", 0xFC),
            0xFE: SMARTAttribute("FREE_FALL_EVENTS", 0xFE, unit=Units.COUNT),
        },
        notes=[
            "The default SMART attributes are based off relatively common"
            " attributes across manufacturers. The attributes are not"
            " guaranteed to be accurate for all drives. If you find any"
            " discrepancies, please open an issue on GitHub."
        ],
    ),
    DriveEntry(
        name="Samsung SSDs",
        filters=[
            "type:ata",
            re.compile(r"model:Samsung SSD 8[56]0 EVO [12]TB"),
        ],
        smart_attributes={
            0xEB: SMARTAttribute("POR_RECOVERY_COUNT", 0xEB, unit=Units.COUNT),
        },
        notes=[
            "Tested against Samsung SSD 850 EVO 2TB",
            "Tested against Samsung SSD 860 EVO 1TB",
        ],
    ),
]


def _match(entry: DriveEntry, filters: List[str]) -> bool:
    """
    Check if the given entry matches the given filters.
    """
    for filter_ in entry.filters:
        if isinstance(filter_, re.Pattern):
            if not any(filter_.match(attr) for attr in filters):
                return False
        elif filter_ not in filters:
            return False

    return True


def get_matching_drive_entries(filters: List[str]) -> List[DriveEntry]:
    """
    Get a list of DriveEntry that matches the given filters.
    """
    return [entry for entry in DRIVE_DATABASE if _match(entry, filters)]


def get_drive_entry(filters: List[str]) -> DriveEntry:
    """
    Get a merged DriveEntry that matches all the given filters.
    """
    entry = DriveEntry(name="_merged_")
    entry.filters = filters

    for matching_entry in get_matching_drive_entries(filters):
        entry.smart_attributes.update(matching_entry.smart_attributes)

    return entry



from pytest_mock import MockerFixture

from tests.scsi.utils import MockedSCSIDevice

VALID_IDENTIFY_RESPONSE = bytes(
    [
        0x40,
        0x00,
        0xFF,
        0x3F,
        0x37,
        0xC8,
        0x10,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x3F,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x33,
        0x51,
        0x21,
        0x5A,
        0x42,
        0x4E,
        0x4B,
        0x30,
        0x32,
        0x33,
        0x33,
        0x35,
        0x32,
        0x38,
        0x20,
        0x54,
        0x20,
        0x20,
        0x20,
        0x20,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x56,
        0x52,
        0x30,
        0x54,
        0x42,
        0x31,
        0x51,
        0x36,
        0x61,
        0x53,
        0x73,
        0x6D,
        0x6E,
        0x75,
        0x20,
        0x67,
        0x53,
        0x53,
        0x20,
        0x44,
        0x36,
        0x38,
        0x20,
        0x30,
        0x56,
        0x45,
        0x20,
        0x4F,
        0x54,
        0x31,
        0x20,
        0x42,
        0x20,
        0x20,
        0x20,
        0x20,
        0x20,
        0x20,
        0x20,
        0x20,
        0x20,
        0x20,
        0x20,
        0x20,
        0x20,
        0x20,
        0x20,
        0x20,
    ]
)


def test_identify(mocker: MockerFixture):
    """
    Test we're properly parsing the response to an SCSI/ATA IDENTIFY.
    """
    get_device = mocker.patch(
        "smartie.device.get_device",
        return_value=MockedSCSIDevice(
            "/dev/test", return_data=VALID_IDENTIFY_RESPONSE
        ),
    )

    with get_device("/dev/test") as device:
        assert device.model == "Samsung SSD 860 EVO 1TB"
        assert device.serial == "Q3Z!NB0K325382T"



"""
Platform-specific utilities such as library imports and external API
definitions.

.. note::

    The individual files are typically not included in documentation as
    they have import side effects and are platform specific.
"""



"""
Constants used when working with NVMe devices.
"""

#: The known possible status field values for an NVMe command.
NVMe_STATUS_FIELD: dict[tuple[int, int], str] = {
    (
        0,
        0,
    ): "Successful Completion: The command completed without error.",
    (
        0,
        0x01,
    ): (
        "Invalid Command Opcode: A reserved coded value or an"
        " unsupported value in the command opcode field."
    ),
    (
        0,
        0x02,
    ): (
        "Invalid Field in Command: A reserved coded value or an"
        " unsupported value in a defined field (other than the opcode"
        " field). This status code should be used unless another status"
        " code is explicitly specified for a particular condition. The"
        " field may be in the command parameters as part of the"
        " Submission Queue Entry or in data structures pointed to by"
        " the command parameters."
    ),
    (
        0,
        0x03,
    ): (
        "Command ID Conflict: The command identifier is already in use."
        " Note: It is implementation specific how many commands are"
        " searched for a conflict."
    ),
    (
        0,
        0x04,
    ): (
        "Data Transfer Error: Transferring the data or metadata"
        " associated with a command had an error."
    ),
    (
        0,
        0x05,
    ): (
        "Commands Aborted due to Power Loss Notification: Indicates"
        " that the command was aborted due to a power loss"
        " notification."
    ),
    (
        0,
        0x06,
    ): (
        "Internal Error: The command was not completed successfully due"
        " to an internal error. Details on the internal device error"
        " should be reported as an asynchronous event. Refer to Figure"
        " 147 for Internal Error Asynchronous Event Information."
    ),
    (
        0,
        0x07,
    ): (
        "Command Abort Requested: The command was aborted due to an"
        " Abort command being received that specified the Submission"
        " Queue Identifier and Command Identifier of this command."
    ),
    (
        0,
        0x08,
    ): (
        "Command Aborted due to SQ Deletion: The command was aborted"
        " due to a Delete I/O Submission Queue request received for the"
        " Submission Queue to which the command was submitted."
    ),
    (
        0,
        0x09,
    ): (
        "Command Aborted due to Failed Fused Command: The command was"
        " aborted due to the other command in a fused operation"
        " failing."
    ),
    (
        0,
        0x0A,
    ): (
        "Command Aborted due to Missing Fused Command: The fused"
        " command was aborted due to the adjacent submission queue"
        " entry not containing a fused command that is the other"
        " command in a supported fused operation."
    ),
    (
        0,
        0x0B,
    ): (
        "Invalid Namespace or Format: The namespace or the format of"
        " that namespace is invalid."
    ),
    (
        0,
        0x0C,
    ): (
        "Command Sequence Error: The command was aborted due to a"
        " protocol violation in a multi-command sequence (e.g., a"
        " violation of the Security Send and Security Receive"
        " sequencing rules in the TCG Storage Synchronous Interface"
        " Communications protocol (refer to TCG Storage Architecture"
        " Core Specification))."
    ),
    (
        0,
        0x0D,
    ): (
        "Invalid SGL Segment Descriptor: The command includes an"
        " invalid SGL Last Segment or SGL Segment descriptor. This may"
        " occur under various conditions."
    ),
    (
        0,
        0x0E,
    ): (
        "Invalid Number of SGL Descriptors: There is an SGL Last"
        " Segment descriptor or an SGL Segment descriptor in a location"
        " other than the last descriptor of a segment based on the"
        " length indicated. This is also used for invalid SGLs in a"
        " command capsule (refer to NVMe over Fabrics specification)."
    ),
    (
        0,
        0x0F,
    ): (
        "Data SGL Length Invalid: This may occur if the length of a"
        " Data SGL is too short. This may occur if the length of a Data"
        " SGL is too long and the controller does not support SGL"
        " transfers longer than the amount of data to be transferred as"
        " indicated in the SGL Support field of the Identify Controller"
        " data structure."
    ),
    (
        0,
        0x10,
    ): (
        "Metadata SGL Length Invalid: This may occur if the length of a"
        " Metadata SGL is too short. This may occur if the length of a"
        " Metadata SGL is too long and the controller does not support"
        " SGL transfers longer than the amount of data to be"
        " transferred as indicated in the SGL Support field of the"
        " Identify Controller data structure."
    ),
    (
        0,
        0x11,
    ): (
        "SGL Descriptor Type Invalid: The type of an SGL Descriptor is"
        " a type that is not supported by the controller, or the"
        " combination of type and subtype is not supported by the"
        " controller."
    ),
    (
        0,
        0x12,
    ): (
        "Invalid Use of Controller Memory Buffer: The attempted use of"
        " the Controller Memory Buffer is not supported by the"
        " controller."
    ),
    (
        0,
        0x13,
    ): (
        "PRP Offset Invalid: The Offset field for a PRP entry is"
        " invalid. This may occur when there is a PRP entry with a"
        " non-zero offset after the first entry or when the Offset"
        " field in any PRP entry is not dword aligned"
    ),
    (
        0,
        0x14,
    ): (
        "Atomic Write Unit Exceeded: The length specified exceeds the"
        " atomic write unit size."
    ),
    (
        0,
        0x15,
    ): (
        "Operation Denied: The command was denied due to lack of access"
        " rights. Refer to the appropriate security specification"
        " (e.g., TCG Storage Interface Interactions Specification). For"
        " media access commands, the Access Denied status code should"
        " be used instead."
    ),
    (
        0,
        0x16,
    ): (
        "SGL Offset Invalid: The offset specified in a descriptor is"
        " invalid. This may occur when using capsules for data"
        " transfers in NVMe over Fabrics implementations and an invalid"
        " offset in the capsule is specified."
    ),
    (0, 0x17): "Reserved",
    (
        0,
        0x18,
    ): (
        "Host Identifier Inconsistent Format: The NVM subsystem"
        " detected the simultaneous use of 64-bit and 128-bit Host"
        " Identifier values on different controllers."
    ),
    (
        0,
        0x19,
    ): "Keep Alive Timer Expired: The Keep Alive Timer expired.",
    (
        0,
        0x1A,
    ): (
        "Keep Alive Timeout Invalid: The Keep Alive Timeout value"
        " specified is invalid. This may be due to an attempt to"
        " specify a value of 0h on a transport that requires the Keep"
        " Alive feature to be enabled. This may be due to the value"
        " specified being too large for the associated NVMe Transport"
        " as defined in the NVMe Transport binding specification."
    ),
    (
        0,
        0x1B,
    ): (
        "Command Aborted due to Preempt and Abort: The command was"
        " aborted due to a Reservation Acquire command with the"
        " Reservation Acquire Action (RACQA) set to 010b (Preempt and"
        " Abort)."
    ),
    (
        0,
        0x1C,
    ): (
        "Sanitize Failed: The most recent sanitize operation failed and"
        " no recovery action has been successfully completed."
    ),
    (
        0,
        0x1D,
    ): (
        "Sanitize In Progress: The requested function (e.g., command)"
        " is prohibited while a sanitize operation is in progress."
    ),
    (
        0,
        0x1E,
    ): (
        "SGL Data Block Granularity Invalid: The Address alignment or"
        " Length granularity for an SGL Data Block descriptor is"
        " invalid. This may occur when a controller supports dword"
        " granularity only and the lower two bits of the Address or"
        " Length are not cleared to 00b."
    ),
    (
        0,
        0x1F,
    ): (
        "Command Not Supported for Queue in CMB: The implementation"
        " does not support submission of the command to a Submission"
        " Queue in the Controller Memory Buffer or command completion"
        " to a Completion Queue in the Controller Memory Buffer."
    ),
    (
        0,
        0x20,
    ): (
        "Namespace is Write Protected: The command is prohibited while"
        " the namespace is write protected as a result of a change in"
        " the namespace write protection state as defined by the"
        " Namespace Write Protection State Machine"
    ),
    (
        0,
        0x21,
    ): (
        "Command Interrupted: Command processing was interrupted and"
        " the controller is unable to successfully complete the"
        " command. The host should retry the command."
    ),
    (
        0,
        0x22,
    ): (
        "Transient Transport Error: A transient transport error was"
        " detected. If the command is retried on the same controller,"
        " the command is likely to succeed. A command that fails with a"
        " transient transport error four or more times should be"
        " treated as a persistent transport error that is not likely to"
        " succeed if retried on the same controller."
    ),
    (
        0,
        0x80,
    ): (
        "LBA Out of Range: The command references an LBA that exceeds"
        " the size of the namespace."
    ),
    (
        0,
        0x81,
    ): (
        "Capacity Exceeded: Execution of the command has caused the"
        " capacity of the namespace to be exceeded. This error occurs"
        " when the Namespace Utilization exceeds the Namespace"
        " Capacity."
    ),
    (
        0,
        0x82,
    ): (
        "Namespace Not Ready: The namespace is not ready to be accessed"
        " as a result of a condition other than a condition that is"
        " reported as an Asymmetric Namespace Access condition. The Do"
        " Not Retry bit indicates whether re-issuing the command at a"
        " later time may succeed."
    ),
    (
        0,
        0x83,
    ): (
        "Reservation Conflict: The command was aborted due to a"
        " conflict with a reservation held on the accessed namespace."
    ),
    (
        0,
        0x84,
    ): (
        "Format In Progress: A Format NVM command is in progress on the"
        " namespace. The Do Not Retry bit shall be cleared to '0' to"
        " indicate that the command may succeed if resubmitted."
    ),
    (0x01, 0): "Completion Queue Invalid.",
    (0x01, 0x01): "Invalid Queue Identifier.",
    (0x01, 0x02): "Invalid Queue Size.",
    (0x01, 0x03): "Abort Command Limit Exceeded.",
    (0x01, 0x04): "Reserved.",
    (0x01, 0x05): "Asynchronous Event Request Limit Exceeded.",
    (0x01, 0x06): "Invalid Firmware Slot.",
    (0x01, 0x07): "Invalid Firmware Image.",
    (0x01, 0x08): "Invalid Interrupt Vector.",
    (0x01, 0x09): "Invalid Log Page.",
    (0x01, 0x0A): "Invalid Format.",
    (0x01, 0x0B): "Firmware Activation Requires Conventional Reset.",
    (0x01, 0x0C): "Invalid Queue Deletion.",
    (0x01, 0x0D): "Feature Identifier Not Saveable.",
    (0x01, 0x0E): "Feature Not Changeable.",
    (0x01, 0x0F): "Feature Not Namespace Specific.",
    (0x01, 0x10): "Firmware Activation Requires NVM Subsystem Reset.",
    (
        0x01,
        0x11,
    ): "Firmware Activation Requires Controller Level Reset.",
    (
        0x01,
        0x12,
    ): "Firmware Activation Requires Maximum Time Violation.",
    (0x01, 0x13): "Firmware Activation Prohibited",
    (0x01, 0x14): "Overlapping Range",
    (0x01, 0x15): "Namespace Insufficient Capacity",
    (0x01, 0x16): "Namespace Identifier Unavailable.",
    (0x01, 0x17): "Reserved.",
    (0x01, 0x18): "Namespace Already Attached.",
    (0x01, 0x19): "Namespace Is Private.",
    (0x01, 0x1A): "Namespace Not Attached.",
    (0x01, 0x1B): "Thin Provisioning Not Supported.",
    (0x01, 0x1C): "Controller List Invalid.",
    (0x01, 0x1D): "Device Self-test In Progress.",
    (0x01, 0x1E): "Boot Partition Write Prohibited.",
    (0x01, 0x1F): "Invalid Controller Identifier.",
    (0x01, 0x20): "Invalid Secondary Controller State.",
    (0x01, 0x21): "Invalid Number of Controller Resources.",
    (0x01, 0x22): "Invalid Resource Identifier.",
    (
        0x01,
        0x23,
    ): "Sanitize Prohibited While Persistent Memory Region is Enabled.",
    (0x01, 0x24): "ANA Group Identifier Invalid.",
    (0x01, 0x25): "ANA Attach Failed.",
    (0x01, 0x80): "Conflicting Attributes.",
    (0x01, 0x81): "Invalid Protection Information.",
    (0x01, 0x82): "Attempted Write to Read Only Range.",
    (
        0x02,
        0x80,
    ): ("Write Fault: The write data could not be committed to the" " media."),
    (
        0x02,
        0x81,
    ): (
        "Unrecovered Read Error: The read data could not be recovered"
        " from the media."
    ),
    (
        0x02,
        0x82,
    ): (
        "End-to-end Guard Check Error: The command was aborted due to"
        " an end-to-end guard check failure."
    ),
    (
        0x02,
        0x83,
    ): (
        "End-to-end Application Tag Check Error: The command was"
        " aborted due to an end-to-end application tag check failure."
    ),
    (
        0x02,
        0x84,
    ): (
        "End-to-end Reference Tag Check Error: The command was aborted"
        " due to an end-to-end reference tag check failure."
    ),
    (
        0x02,
        0x85,
    ): (
        "Compare Failure: The command failed due to a miscompare during"
        " a Compare command."
    ),
    (
        0x02,
        0x86,
    ): (
        "Access Denied: Access to the namespace and/or LBA range is"
        " denied due to lack of access rights. Refer to the appropriate"
        " security specification (e.g., TCG Storage Interface"
        " Interactions Specification)."
    ),
    (
        0x02,
        0x87,
    ): (
        "Deallocated or Unwritten Logical Block: The command failed due"
        " to an attempt to read from or verify an LBA range containing"
        " a deallocated or unwritten logical block"
    ),
    (
        0x03,
        0,
    ): (
        "Internal Path Error: The command was not completed as the"
        " result of a controller internal error that is specific to the"
        " controller processing the command. Retries for the request"
        " function should be based on the setting of the DNR bit"
    ),
    (
        0x03,
        0x01,
    ): (
        "Asymmetric Access Persistent Loss: The requested function"
        " (e.g., command) is not able to be performed as a result of"
        " the relationship between the controller and the namespace"
        " being in the ANA Persistent Loss state (refer to section"
        " 8.20.3.4). The command should not be re-submitted to the same"
        " controller"
    ),
    (
        0x03,
        0x02,
    ): (
        "Asymmetric Access Inaccessible: The requested function (e.g.,"
        " command) is not able to be performed as a result of the"
        " relationship between the controller and the namespace being"
        " in the ANA Inaccessible state (refer to section 8.20.3.3)."
        " The command should not be re-submitted to the same"
        " controller."
    ),
    (
        0x03,
        0x03,
    ): (
        "Asymmetric Access Transition: The requested function (e.g.,"
        " command) is not able to be performed as a result of the"
        " relationship between the controller and the namespace"
        " transitioning between Asymmetric Namespace Access states"
        " (refer to section 8.20.3.5). The requested function should be"
        " retried after the transition is complete."
    ),
    (
        0x03,
        0x60,
    ): (
        "Controller Pathing Error: A pathing error was detected by the"
        " controller."
    ),
    (
        0x03,
        0x70,
    ): "Host Pathing Error: A pathing error was detected by the host.",
    (
        0x03,
        0x71,
    ): (
        "Command Aborted By Host: The command was aborted as a result"
        " of host action (e.g., the host disconnected the Fabric"
        " connection)."
    ),
}



"""
This file contains the various low-level structure definitions used for sending
and receiving NVMe commands, as well as the structures required for
platform-specific APIs.
"""

import ctypes
import enum

from smartie.structures import c_uint128

#: IOCTL for NVMe Admin commands on Linux.
IOCTL_NVMe_ADMIN_CMD = 0xC0484E41


class NVMeAdminCommands(enum.IntEnum):
    GET_LOG_PAGE = 0x02
    IDENTIFY = 0x06


class NVMeAdminCommand(ctypes.Structure):
    _pack_ = 1
    _fields_ = [
        ("opcode", ctypes.c_ubyte),
        ("flags", ctypes.c_ubyte),
        ("reserved_1", ctypes.c_ushort),
        ("nsid", ctypes.c_uint32),
        ("cdw2", ctypes.c_uint32),
        ("cdw3", ctypes.c_uint32),
        ("metadata", ctypes.c_uint64),
        ("addr", ctypes.c_uint64),
        ("metadata_len", ctypes.c_uint32),
        ("data_len", ctypes.c_uint32),
        ("cdw10", ctypes.c_uint32),
        ("cdw11", ctypes.c_uint32),
        ("cdw12", ctypes.c_uint32),
        ("cdw13", ctypes.c_uint32),
        ("cdw14", ctypes.c_uint32),
        ("cdw15", ctypes.c_uint32),
        ("timeout_ms", ctypes.c_uint32),
        ("result", ctypes.c_uint32),
    ]


class NVMeCQEStatusField(ctypes.Structure):
    """
    The format of NVMe Completion Queue Entry Status Field
    """

    _fields_ = [
        ("status_code", ctypes.c_ubyte),
        ("status_code_type", ctypes.c_ubyte, 3),
        ("cmd_retry_delay", ctypes.c_ubyte, 2),
        ("more", ctypes.c_ubyte, 1),
        ("do_not_retry", ctypes.c_ubyte, 1),
    ]


class NVMeIdentifyResponse(ctypes.Structure):
    _fields_ = [
        ("vendor_id", ctypes.c_uint16),
        ("ssvid", ctypes.c_uint16),
        ("serial_number", ctypes.c_ubyte * 20),
        ("model_number", ctypes.c_ubyte * 40),
        ("firmware_revision", ctypes.c_ubyte * 8),
        # The majority of this structure has yet to be implemented. Add fields
        # as ya need em.
        ("unknown", ctypes.c_ubyte * 4024),
    ]


class SMARTCriticalWarning(ctypes.Structure):
    _pack_ = 1
    _fields_ = [
        ("available_spare", ctypes.c_ubyte, 1),
        ("temperature", ctypes.c_ubyte, 1),
        ("degraded_nvm", ctypes.c_ubyte, 1),
        ("read_only", ctypes.c_ubyte, 1),
        ("volatile_memory_backup", ctypes.c_ubyte, 1),
        ("reserved_1", ctypes.c_ubyte, 3),
    ]


class SMARTPageResponse(ctypes.Structure):
    """
    This structure represents the response from the SMART Log Page (0x02).

    .. note::

        Defined in the NVMe 1.4 specification as figure 194.
    """

    _pack_ = 1
    _fields_ = [
        # Flags for any critical warnings.
        ("critical_warning", SMARTCriticalWarning),
        # A composite of the temperature in Kelvin. How this is calculated is
        # not really defined, and may be an average of multiple sensors.
        ("temperature", ctypes.c_ushort),
        ("available_spare", ctypes.c_ubyte),
        ("available_spare_threshold", ctypes.c_ubyte),
        ("percent_used", ctypes.c_ubyte),
        ("endurance_group_critical_warning_summary", ctypes.c_ubyte),
        ("reserved_1", ctypes.c_ubyte * 25),
        # The number of 512-byte data units read, in thousands.
        ("data_units_read", c_uint128),
        # The number of 512-byte data units written, in thousands.
        ("data_units_written", c_uint128),
        ("host_read_commands", c_uint128),
        ("host_write_commands", c_uint128),
        ("controller_busy_time", c_uint128),
        ("power_cycles", c_uint128),
        ("power_on_hours", c_uint128),
        ("unsafe_shutdowns", c_uint128),
        ("media_errors", c_uint128),
        ("num_err_log_entries", c_uint128),
        ("warning_temp_time", ctypes.c_uint32),
        ("critical_temp_time", ctypes.c_uint32),
        ("temperature_sensors", ctypes.c_uint16 * 8),
        ("thermal_transition_counts", ctypes.c_uint32 * 2),
        ("total_time_for_thermal_management", ctypes.c_uint32 * 2),
        ("reserved_2", ctypes.c_ubyte * 280),
    ]



class SenseError(Exception):
    def __init__(self, error_code: int, *, sense=None):
        super().__init__()
        self.error_code = error_code
        self.sense = sense

    @property
    def error_msg(self) -> str:
        return {
            0x00: "No Sense",
            0x01: "Recovered Error",
            0x02: "Not Ready",
            0x03: "Medium Error",
            0x04: "Hardware Error",
            0x05: "Illegal Request",
            0x06: "Unit Attention",
            0x07: "Data Protect",
            0x09: "Firmware Error",
            0x0B: "Aborted Command",
            0x0C: "Equal",
            0x0D: "Volume Overflow",
            0x0E: "Miscompare",
            0x0F: "Completed",
        }.get(self.error_code, "Unknown Sense Error")

    def __repr__(self):
        return (
            f"<{self.__class__.__name__}(error_code=0x{self.error_code:02x},"
            f" err={self.error_msg!r})>"
        )

    __str__ = __repr__



__all__ = ("NVMeDevice", "NVMeResponse", "local_byteorder")
import abc
import sys
import ctypes
from dataclasses import dataclass
from abc import abstractmethod
from typing import Optional, Tuple, Union, Any

from smartie.device import Device
from smartie.nvme.errors import NVMeStatusFieldError
from smartie.nvme.structures import (
    NVMeAdminCommand,
    NVMeAdminCommands,
    NVMeIdentifyResponse,
    NVMeCQEStatusField,
    SMARTPageResponse,
)
from smartie.structures import structure_to_dict

local_byteorder = sys.byteorder


@dataclass
class NVMeResponse:
    """
    Common response object for NVMe commands.

    This object attempts to encapsulate the response from a NVMe command in a
    platform-agnostic way. It contains the Status Field, the command that was
    issued, command spec value and whether the command succeeded or not.

    For additional platform-specific information, the `platform_header`
    attribute contains the platform-specific header that was used to issue the
    command.
    """

    #: Whether the command succeeded. If None, the status is unknown.
    succeeded: Optional[bool]
    #: The Command Specific value returned by the device.
    #: The value maybe either of command error information or
    #: the wanted value return by no data transfer command
    command_spec: Optional[int]
    #: The status field data returned by the device.
    status_field: Optional[NVMeCQEStatusField]
    #: The command issued to the device.
    command: Union[NVMeAdminCommand,]
    #: Keep aligned with SCSIResponse. Not used for now.
    #: The actual number of bytes transferred.
    bytes_transferred: Optional[int]

    #: Keep aligned with SCSIResponse. Not used for now.
    #: The platform-specific header that was used to issue the command.
    #: For example this may be an :class:`SCSIPassThroughDirectWithBuffer` on
    #: Windows.
    platform_header: Any

    def __bool__(self):
        return self.succeeded


class NVMeDevice(Device, abc.ABC):
    @classmethod
    def parse_status_field(cls, status_blob) -> Optional[NVMeCQEStatusField]:
        """
        Parses the command status field from an NVMe command, raising a
        :class:`smartie.nvme.errors.NVMeStatusFieldError` if an error occurred.

        :param status_blob: A bytes/bytearray (or similar) object containing
                            the unparsed sense response.
        """
        status_field = NVMeCQEStatusField.from_buffer_copy(status_blob)
        if status_field.status_code != 0 or status_field.status_code_type != 0:
            raise NVMeStatusFieldError(
                status_field.status_code,
                status_field.status_code_type,
                status_field=status_field,
            )
        return status_field

    @abstractmethod
    def issue_admin_command(self, command):
        pass

    def identify(self) -> Tuple[NVMeIdentifyResponse, NVMeResponse]:
        """
        Returns the parsed IDENTIFY results for CNS 01h, which contains
        the controller information.
        """
        data = NVMeIdentifyResponse()
        response = self.issue_admin_command(
            NVMeAdminCommand(
                opcode=NVMeAdminCommands.IDENTIFY,
                addr=ctypes.addressof(data),
                data_len=ctypes.sizeof(data),
                cdw10=1,
            )
        )
        return data, response

    @property
    def serial(self) -> Optional[str]:
        identify, response = self.identify()
        return bytearray(identify.serial_number).strip(b" \x00").decode()

    @property
    def model(self) -> Optional[str]:
        identify, response = self.identify()
        return bytearray(identify.model_number).strip(b" \x00").decode()

    @property
    def temperature(self) -> Optional[int]:
        smart, response = self.smart()
        return int(smart.temperature - 273.15)

    def read_log_page(
        self, log_page_id: int, data: ctypes.Structure
    ) -> tuple[ctypes.Structure, NVMeResponse]:
        response = self.issue_admin_command(
            NVMeAdminCommand(
                opcode=NVMeAdminCommands.GET_LOG_PAGE,
                addr=ctypes.addressof(data),
                data_len=ctypes.sizeof(data),
                nsid=0xFFFFFFFF,
                cdw10=log_page_id | (((ctypes.sizeof(data) // 4) - 1) << 16),
            )
        )
        return data, response

    def smart(self):
        return self.read_log_page(0x02, SMARTPageResponse())

    @property
    def smart_table(self) -> dict[str, any]:
        smart, response = self.smart()
        return structure_to_dict(smart)



import ctypes
import os
from typing import Union

from smartie.scsi import SCSIDevice, SCSIResponse
from smartie.platforms.linux import get_libc
from smartie.scsi.structures import (
    DescriptorFormatSense,
    FixedFormatSense,
    IOCTL_SG_IO,
    SGIOHeader,
    Direction,
    StatusCode,
)


class LinuxSCSIDevice(SCSIDevice):
    def __enter__(self):
        if self.fd is not None:
            raise IOError("Device already open.")

        self.fd = os.open(self.path, os.O_RDONLY | os.O_NONBLOCK)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.fd is not None:
            os.close(self.fd)
            self.fd = None
        return False

    def issue_command(
        self,
        direction: Direction,
        command: ctypes.Structure,
        data: Union[ctypes.Array, ctypes.Structure, None],
        *,
        timeout: int = 3000,
    ) -> SCSIResponse:
        # The Sense response can be in multiple formats, and we won't know
        # what it is until we see the first byte.
        raw_sense = ctypes.create_string_buffer(
            b"\x00",
            max(
                ctypes.sizeof(FixedFormatSense),
                ctypes.sizeof(DescriptorFormatSense),
            ),
        )

        if data is None:
            data = ctypes.create_string_buffer(0)

        sg_io_header = SGIOHeader(
            interface_id=83,  # Always 'S'
            dxfer_direction=direction,
            cmd_len=ctypes.sizeof(command),
            cmdp=ctypes.addressof(command),
            dxfer_len=ctypes.sizeof(data),
            dxferp=ctypes.addressof(data),
            mx_sb_len=ctypes.sizeof(raw_sense),
            sbp=ctypes.addressof(raw_sense),
            timeout=timeout,
        )

        # We use libc instead of the builtin ioctl as the builtin can have
        # issues with 64-bit pointers.
        result = get_libc().ioctl(
            self.fd, IOCTL_SG_IO, ctypes.byref(sg_io_header)
        )

        if result != 0:
            raise OSError(ctypes.get_errno())

        return SCSIResponse(
            succeeded=(
                sg_io_header.status == StatusCode.GOOD
                or sg_io_header.status == StatusCode.CONDITION_GOOD
            ),
            sense=self.parse_sense(raw_sense.raw),
            platform_header=sg_io_header,
            command=command,
            bytes_transferred=sg_io_header.dxfer_len - sg_io_header.resid,
        )



import ctypes
from typing import Optional, Union

from smartie.scsi import SCSIDevice, SCSIResponse
from smartie.scsi.structures import Direction


class MockedSCSIDevice(SCSIDevice):
    def __init__(
        self,
        *args,
        return_data: Optional[bytes] = None,
        return_sense=None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._return_data = return_data
        self._return_sense = return_sense

    def issue_command(
        self,
        direction: Direction,
        command: ctypes.Structure,
        data: Union[ctypes.Array, ctypes.Structure, None],
        *,
        timeout: int = 3000,
    ) -> SCSIResponse:
        if self._return_data:
            ctypes.memmove(data, self._return_data, len(self._return_data))

        return SCSIResponse(
            sense=self._return_sense,
            platform_header=None,
            bytes_transferred=len(self._return_data)
            if self._return_data
            else 0,
            succeeded=True,
            command=command,
        )

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False



"""
This file contains the various low-level structure definitions used for sending
and receiving SCSI commands, as well as the structures required for
platform-specific APIs.

Where reasonable, the names of fields have been taken from the specifications
to ease reference searches.
"""

import ctypes
import enum


#: The Linux IOCTL for SG_IO, which is the driver we use to send and receive
#: SCSI commands.
IOCTL_SG_IO = 0x2285
#: The Windows IOCTL for SCSI pass-through, which is what we use to send and
#: receive SCSI commands.
IOCTL_SCSI_PASS_THROUGH_DIRECT = 0x4D014


class DeviceType(enum.IntEnum):
    """
    Common device types returned by the SCSI INQUIRY command.

    .. note::

        Not all possible values are defined here due to lack of testable
        hardware.
    """

    #: Direct access block device (ex: disk)
    DIRECT_ACCESS_BLOCK_DEVICE = 0x00
    #: Sequential access device (ex: tape drive)
    SEQUENTIAL_ACCESS_DEVICE = 0x01
    #: CD/DVD/BLU-RAY
    CDROM = 0x05
    #: Some types of optical disks
    OPTICAL_MEMORY_DEVICE = 0x07
    #: Storage array controller device
    RAID_CONTROLLER = 0x0C
    #: Optical card reader/writer device
    OPTICAL_CARD_RW = 0x0F


class OperationCode(enum.IntEnum):
    #: SCSI INQUIRY command.
    INQUIRY = 0x12
    #: 16-byte ATA passthrough command.
    COMMAND_16 = 0x85
    #: 12-byte ATA passthrough command.
    COMMAND_12 = 0xA1


class ATAProtocol(enum.IntEnum):
    """
    The possible values for `Command16.protocol`.
    """

    HARD_RESET = 0
    SRST = 1
    NON_DATA = 3
    PIO_DATA_IN = 4
    PIO_DATA_OUT = 5
    DMA = 6
    DMA_QUEUED = 7
    DEVICE_DIAGNOSTIC = 8
    DEVICE_RESET = 9
    UDMA_DATA_IN = 10
    UDMA_DATA_OUT = 11
    FPDMA = 12
    RETURN_RESPONSE_INFORMATION = 15


class ATACommands(enum.IntEnum):
    """
    The possible values for `Command16.command`.
    """

    SMART = 0xB0
    IDENTIFY = 0xEC
    READ_DATA = 0xD0
    READ_LOG = 0xD5
    RETURN_STATUS = 0xDA


class ATAPICommands(enum.IntEnum):
    """
    The possible values for `Command16.command` when targeting an ATAPI device.
    """

    IDENTIFY = 0xA1


class StatusCode(enum.IntEnum):
    """
    The possible values for `SGIOHeader.status`.
    """

    GOOD = 0x00
    CHECK_CONDITION = 0x01
    CONDITION_GOOD = 0x02
    BUSY = 0x04
    INTERMEDIATE_GOOD = 0x08
    INTERMEDIATE_C_GOOD = 0x0A
    RESERVATION_CONFLICT = 0x0C


class SenseErrorCode(enum.IntEnum):
    CURRENT_ERROR = 0x70
    DEFERRED_ERROR = 0x71


class Direction(enum.IntEnum):
    """
    The direction of the command being sent by a call to
    DeviceIO.issue_command.

    .. note::

        These are really the constants for the direction in SG_IO calls, but we
        map them for other platforms.
    """

    NONE = -1
    TO = -2
    FROM = -3


class ATASmartFeature(enum.IntEnum):
    """
    The possible values for the `feature` field on an ATA SMART command.
    """

    SMART_READ_DATA = 0xD0
    SMART_READ_THRESHOLDS = 0xD1
    SMART_TOGGLE_ATTRIBUTE_AUTOSAVE = 0xD2
    SMART_EXECUTE_OFF_LINE_IMMEDIATE = 0xD4
    SMART_READ_LOG = 0xD5
    SMART_WRITE_LOG = 0xD6
    SMART_ENABLE_OPERATIONS = 0xD8
    SMART_DISABLE_OPERATIONS = 0xD9
    SMART_RETURN_STATUS = 0xDA


class InquiryCommand(ctypes.Structure):
    """
    An SCSI INQUIRY command.
    """

    _pack_ = 1
    _fields_ = [
        ("operation_code", ctypes.c_ubyte),
        ("lun", ctypes.c_ubyte),
        ("page_code", ctypes.c_ubyte),
        ("reserved_1", ctypes.c_ubyte),
        ("allocation_length", ctypes.c_ubyte),
        ("control", ctypes.c_ubyte),
    ]


class InquiryResponse(ctypes.Structure):
    """
    The response to an SCSI INQUIRY command.
    """

    _pack_ = 1
    _fields_ = [
        ("peripheral_device_type", ctypes.c_ubyte, 4),
        ("peripheral_qualifier", ctypes.c_ubyte, 4),
        ("reserved_1", ctypes.c_ubyte, 4),
        ("hot_pluggable", ctypes.c_ubyte, 2),
        ("lu_cong", ctypes.c_ubyte, 1),
        ("rmb", ctypes.c_ubyte, 1),
        ("version", ctypes.c_ubyte),
        ("response_data_format", ctypes.c_ubyte, 4),
        ("hi_sup", ctypes.c_ubyte, 1),
        ("norm_aca", ctypes.c_ubyte, 1),
        ("reserved_2", ctypes.c_ubyte, 1),
        ("reserved_3", ctypes.c_ubyte, 1),
        ("additional_length", ctypes.c_ubyte),
        ("protect", ctypes.c_ubyte, 1),
        ("reserved_4", ctypes.c_ubyte, 2),
        ("three_pc", ctypes.c_ubyte, 1),
        ("tpgs", ctypes.c_ubyte, 1),
        ("obsolete_1", ctypes.c_ubyte, 1),
        ("sccs", ctypes.c_ubyte, 1),
        ("obsolete_2", ctypes.c_ubyte, 1),
        ("reserved_5", ctypes.c_ubyte, 1),
        ("reserved_6", ctypes.c_ubyte, 1),
        ("obsolete_3", ctypes.c_ubyte, 1),
        ("multi_p", ctypes.c_ubyte, 1),
        ("vs_1", ctypes.c_ubyte, 1),
        ("enc_serv", ctypes.c_ubyte, 1),
        ("obsolete_4", ctypes.c_ubyte, 1),
        ("vs_2", ctypes.c_ubyte, 1),
        ("cmd_que", ctypes.c_ubyte, 1),
        ("reserved_7", ctypes.c_ubyte, 1),
        ("obsolete_5", ctypes.c_ubyte, 1),
        ("obsolete_6", ctypes.c_ubyte, 2),
        ("reserved_8", ctypes.c_ubyte, 1),
        ("obsolete_7", ctypes.c_ubyte, 1),
        ("t10_vendor_identification", ctypes.c_ubyte * 8),
        ("product_identification", ctypes.c_ubyte * 16),
        ("product_revision_level", ctypes.c_ubyte * 4),
        ("vendor_specific_1", ctypes.c_ubyte * 20),
        ("obsolete_8", ctypes.c_ubyte, 4),
        ("reserved_9", ctypes.c_ubyte, 4),
        ("reserved_10", ctypes.c_ubyte),
        ("version_descriptors", ctypes.c_ushort * 16),
        ("reserved_11", ctypes.c_ubyte * 22),
    ]


class CommandFlags(ctypes.Structure):
    """
    The flags used in the SCSI Command16 and Command12 structures.
    """

    class Length(enum.IntEnum):
        """
        Possible values for the `t_length` field.
        """

        #: The transfer length is in the FEATURE field.
        IN_FEATURE = 0b01
        #: The transfer length is in the SECTOR_COUNT field.
        IN_SECTOR_COUNT = 0b10
        #: The transfer length is in the STPIU.
        IN_STPSIU = 0b11

    class OffLine(enum.IntFlag):
        """
        Possible values for the `off_line` field.
        """

        ZERO_SECONDS = 0b00
        TWO_SECONDS = 0b01
        SIX_SECONDS = 0b10
        FOURTEEN_SECONDS = 0b11

    _pack_ = 1
    _fields_ = [
        # If set, determines how the transfer length is specified.
        ("t_length", ctypes.c_ubyte, 2),
        # 0 if the t_length is in bytes, 1 if the t_length is in blocks.
        # Ignored if t_length is 0.
        ("byt_blok", ctypes.c_ubyte, 1),
        # If set to 0, the transfer is from the client to the ATA device.
        # If set to 1, the transfer is from the ATA device to the client.
        ("t_dir", ctypes.c_ubyte, 1),
        ("reserved_1", ctypes.c_ubyte, 1),
        # If set, the ATA device will copy the ATA register information to the
        # sense data even when no error occurs.
        ("ck_cond", ctypes.c_ubyte, 1),
        # The number of seconds to wait for the device to become ready after
        # issuing an ATA command to a PATA device that may cause it to be in
        # an unusable state.
        ("off_line", ctypes.c_ubyte, 2),
    ]


class Command12(ctypes.Structure):
    """
    A 12-byte SCSI/ATA passthrough command.

    .. note::

        The contents of this structure are documented in 04-262r8.pdf, section
        13.2.2.
    """

    _pack_ = 1
    _fields_ = [
        ("operation_code", ctypes.c_ubyte),
        ("protocol", ctypes.c_ubyte),
        ("flags", CommandFlags),
        ("features", ctypes.c_ubyte),
        ("reserved_1", ctypes.c_ubyte, 1),
        ("sector_count", ctypes.c_ubyte, 7),
        ("reserved_2", ctypes.c_ubyte, 1),
        ("lba_low", ctypes.c_ubyte, 7),
        ("reserved_3", ctypes.c_ubyte, 1),
        ("lba_mid", ctypes.c_ubyte, 7),
        ("reserved_4", ctypes.c_ubyte, 1),
        ("lba_high", ctypes.c_ubyte, 7),
        ("device", ctypes.c_ubyte),
        ("command", ctypes.c_ubyte),
        ("reserved_5", ctypes.c_ubyte),
        ("control", ctypes.c_ubyte),
    ]


class Command16(ctypes.Structure):
    """
    A 16-byte SCSI/ATA passthrough command.

    .. note::

        The contents of this structure are documented in 04-262r8.pdf, section
        13.2.3.
    """

    _pack_ = 1
    _fields_ = [
        ("operation_code", ctypes.c_ubyte),
        ("protocol", ctypes.c_ubyte),
        ("flags", CommandFlags),
        ("features", ctypes.c_ushort),
        ("sector_count", ctypes.c_ushort),
        ("lba_high_low", ctypes.c_ubyte),
        ("lba_low", ctypes.c_ubyte),
        ("lba_high_mid", ctypes.c_ubyte),
        ("lba_mid", ctypes.c_ubyte),
        ("lba_high_high", ctypes.c_ubyte),
        ("lba_high", ctypes.c_ubyte),
        ("device", ctypes.c_ubyte),
        ("command", ctypes.c_ubyte),
        ("control", ctypes.c_ubyte),
    ]

    def set_lba(self, lba: int):
        lba = lba.to_bytes(6, byteorder="little")
        self.lba_high_low = lba[3]
        self.lba_low = lba[0]
        self.lba_high_mid = lba[4]
        self.lba_mid = lba[1]
        self.lba_high_high = lba[5]
        self.lba_high = lba[2]

        return self


class SGIOHeader(ctypes.Structure):
    """
    Corresponds to the compat_sg_io_hdr structure in <scsi/sg.h> on Linux.
    """

    _fields_ = [
        ("interface_id", ctypes.c_int),
        ("dxfer_direction", ctypes.c_int),
        ("cmd_len", ctypes.c_ubyte),
        ("mx_sb_len", ctypes.c_ubyte),
        ("iovec_count", ctypes.c_ushort),
        ("dxfer_len", ctypes.c_uint),
        ("dxferp", ctypes.c_void_p),
        ("cmdp", ctypes.c_void_p),
        ("sbp", ctypes.c_void_p),
        ("timeout", ctypes.c_uint),
        ("flags", ctypes.c_uint),
        ("pack_id", ctypes.c_int),
        ("usr_ptr", ctypes.c_void_p),
        ("status", ctypes.c_ubyte),
        ("masked_status", ctypes.c_ubyte),
        ("msg_status", ctypes.c_ubyte),
        ("sb_len_wr", ctypes.c_ubyte),
        ("host_status", ctypes.c_ushort),
        ("driver_status", ctypes.c_ushort),
        ("resid", ctypes.c_int),
        ("duration", ctypes.c_uint),
        ("info", ctypes.c_uint),
    ]


class SCSIPassThroughDirect(ctypes.Structure):
    """
    Corresponds to the SCSI_PASS_THROUGH_DIRECT structure in <ntddscsi.h> on
    Windows.
    """

    _fields_ = [
        ("length", ctypes.c_ushort),
        ("scsi_status", ctypes.c_ubyte),
        ("path_id", ctypes.c_ubyte),
        ("target_id", ctypes.c_ubyte),
        ("lun", ctypes.c_ubyte),
        ("cdb_length", ctypes.c_ubyte),
        ("sense_info_length", ctypes.c_ubyte),
        ("data_in", ctypes.c_ubyte),
        ("data_transfer_length", ctypes.c_uint32),
        ("timeout_value", ctypes.c_uint32),
        ("data_buffer", ctypes.c_void_p),
        ("sense_info_offset", ctypes.c_uint32),
        ("cdb", ctypes.c_ubyte * 16),
    ]


class SCSIPassThroughDirectWithBuffer(ctypes.Structure):
    """
    Corresponds to the SCSI_PASS_THROUGH_DIRECT_WITH_BUFFER structure in
    <ntddscsi.h> on Windows.
    """

    _fields_ = [
        ("sptd", SCSIPassThroughDirect),
        ("filler", ctypes.c_uint32),
        ("sense", ctypes.c_ubyte * 32),
    ]


class FixedFormatSense(ctypes.Structure):
    """
    The SENSE response may be in of two formats - this one, or
    :class:`DescriptorFormatSense`. The exact format depends on the value of
    the first byte, `error_code`.
    """

    _fields_ = [
        ("error_code", ctypes.c_ubyte, 7),
        ("valid", ctypes.c_ubyte, 1),
        ("segment_number", ctypes.c_ubyte),
        ("sense_key", ctypes.c_ubyte, 4),
        ("reserved_1", ctypes.c_ubyte, 1),
        ("ili", ctypes.c_ubyte, 1),
        ("eom", ctypes.c_ubyte, 1),
        ("filemark", ctypes.c_ubyte, 1),
        ("information", ctypes.c_uint32),
        ("additional_sense_length", ctypes.c_ubyte),
        ("command_specific_information", ctypes.c_uint32),
        ("additional_sense_code", ctypes.c_ubyte),
        ("additional_sense_code_qualifier", ctypes.c_ubyte),
        ("field_replaceable_unit_code", ctypes.c_ubyte),
        ("sense_key_specific", ctypes.c_ubyte * 3),
    ]


class DescriptorFormatSense(ctypes.Structure):
    """
    The SENSE response may be in of two formats - this one, or
    :class:`FixedFormatSense`. The exact format depends on the value of
    the first byte, `error_code`.
    """

    _fields_ = [
        ("error_code", ctypes.c_ubyte, 7),
        ("valid", ctypes.c_ubyte, 1),
        ("sense_key", ctypes.c_ubyte, 4),
        ("reserved_1", ctypes.c_ubyte, 4),
        ("additional_sense_code", ctypes.c_ubyte),
        ("additional_sense_code_qualifier", ctypes.c_ubyte),
    ]


class IdentifyResponse(ctypes.Structure):
    """
    The response to an SCSI/ATA IDENTIFY command.

    .. note::

        This is a large structure, and has only been partially implemented. The
        full response is 512 bytes.
    """

    # I don't have the willpower to implement this entire structure. If you
    # need a field, add it.
    _fields_ = [
        # ('general_config', ctypes.c_ushort),
        ("reserved_1", ctypes.c_ushort, 1),
        ("retired_3", ctypes.c_ushort, 1),
        ("response_incomplete", ctypes.c_ushort, 1),
        ("retired_2", ctypes.c_ushort, 3),
        ("fixed_device", ctypes.c_ushort, 1),
        ("removable_media", ctypes.c_ushort, 1),
        ("retired_1", ctypes.c_ushort, 7),
        ("device_type", ctypes.c_ushort, 1),
        # We don't use the following 18 bytes.
        ("padding_1", ctypes.c_ubyte * 18),
        ("serial_number", ctypes.c_ubyte * 20),
        # We don't use the following 6 bytes.
        ("padding_2", ctypes.c_ubyte * 6),
        ("firmware_revision", ctypes.c_ubyte * 8),
        ("model_number", ctypes.c_ubyte * 40),
    ]


class SmartDataEntry(ctypes.Structure):
    """
    An entry in the SMART attribute table.

    .. note::

        The specification calls this field vendor specific, but its format is
        very consistent.
    """

    _pack_ = 1
    _fields_ = [
        ("id", ctypes.c_ubyte),
        ("flags", ctypes.c_ushort),
        ("current", ctypes.c_ubyte),
        ("worst", ctypes.c_ubyte),
        ("vendor_specific_1", ctypes.c_ubyte * 6),
        ("reserved", ctypes.c_ubyte),
    ]


class SmartDataResponse(ctypes.Structure):
    """
    The result of a SMART READ_DATA command.

    .. note::

        The most interesting field in here is likely `vendor_specific_1`,
        which contains the SMART attribute table that encodes values such as
        the current device temperature.

        Although the specification calls this field vendor specific, its
        format is very consistent. See
        :func:`smartie.smart.parse_smart_read_data()` for an example.
    """

    _fields_ = [
        ("version", ctypes.c_ushort),
        ("attributes", SmartDataEntry * 30),
        ("offline_data_collection_status", ctypes.c_ubyte),
        ("self_test_execution_status_buyte", ctypes.c_ubyte),
        ("vendor_specific_2", ctypes.c_ubyte * 2),
        ("vendor_specific_3", ctypes.c_ubyte),
        ("offline_data_collection_capability", ctypes.c_ubyte),
        ("smart_capability", ctypes.c_ushort),
        ("error_logging_capability", ctypes.c_ubyte),
        ("vendor_specific_4", ctypes.c_ubyte),
        # Recommended time is in minutes.
        ("short_self_test_recommended_time", ctypes.c_ubyte),
        # Recommended time is in minutes.
        ("extended_self_test_recommended_time", ctypes.c_ubyte),
        # Recommended time is in minutes.
        ("conveyance_self_test_recommended_time", ctypes.c_ubyte),
        # Recommended time is in minutes.
        ("extended_self_test_recommended_time_wide", ctypes.c_short),
        ("reserved_1", ctypes.c_ubyte * 8),
        ("vendor_specific_2", ctypes.c_ubyte * 124),
        ("data_checksum_structure", ctypes.c_ubyte),
    ]


class SmartThresholdEntry(ctypes.Structure):
    """
    A single entry in the SMART READ_THRESHOLDS response.
    """

    _fields_ = [
        ("attribute_id", ctypes.c_ubyte),
        ("value", ctypes.c_ubyte),
        ("reserved_1", ctypes.c_ubyte * 10),
    ]


class SmartThresholdResponse(ctypes.Structure):
    """
    The result of a SMART READ_THRESHOLDS command.

    .. note::
        The most interesting field in here is likely `vendor_specific_1`,
        which contains the SMART threshold table that encodes values such as
        the temperature at which the device will start throttling.
    """

    _fields_ = [
        ("revision_number", ctypes.c_ushort),
        ("entries", SmartThresholdEntry * 30),
        ("reserved_1", ctypes.c_ubyte * 149),
        ("checksum", ctypes.c_ubyte),
    ]



import ctypes
import os

from smartie.nvme import (
    NVMeDevice,
    NVMeResponse,
    local_byteorder,
)
from smartie.platforms.linux import get_libc
from smartie.nvme.structures import IOCTL_NVMe_ADMIN_CMD, NVMeAdminCommand


class LinuxNVMeDevice(NVMeDevice):
    """
    Represents an NVMe device on a Linux system.
    """

    def __enter__(self):
        if self.fd is not None:
            raise IOError("Device already open.")

        self.fd = os.open(self.path, os.O_RDONLY | os.O_NONBLOCK)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.fd is not None:
            os.close(self.fd)
            self.fd = None
        return False

    def issue_admin_command(self, command: NVMeAdminCommand) -> NVMeResponse:
        result = get_libc().ioctl(
            self.fd, IOCTL_NVMe_ADMIN_CMD, ctypes.byref(command)
        )

        # Most commonly because the caller doesn't have root.
        if result < 0:
            raise OSError(f"NVMe Admin command failed with error {result}")

        # Status is in the upper 16 bits of the result
        status_field = self.parse_status_field(
            command.result.to_bytes(2, local_byteorder)
        )

        return NVMeResponse(
            succeeded=status_field.status_code == 0,
            command_spec=command.result,
            status_field=status_field,
            command=command,
            bytes_transferred=None,
            platform_header=None,
        )



"""
High-level abstractions for enumerating devices and getting basic device
information.
"""

import abc
import itertools
import ctypes
import platform
import os
from pathlib import Path
from typing import Iterable, List, Optional, Union


class Device(abc.ABC):
    """
    A Device represents a high-level abstraction over a system block
    device.


    .. note::

        Typically, an end user will never instantiate a Device directly, but
        instead use the :meth:`get_device()` function to get a `Device`
        instance for a given path. This function will automatically determine
        the correct subclass to use based on the platform and device type.

    :param path: The filesystem path to the device (such as /dev/sda).
    """

    path: str
    fd: Optional[int]

    def __init__(self, path: Union[Path, str]):
        self.path = str(path)
        self.fd = None

    def __repr__(self):
        return f"{self.__class__.__name__}(path={self.path!r})"

    @property
    def model(self) -> Optional[str]:
        """
        Returns the model name of the device.
        """
        return None

    @property
    def serial(self) -> Optional[str]:
        """
        Returns the serial number of the device.
        """
        return None

    @property
    def temperature(self) -> Optional[int]:
        """
        Returns the temperature of the device in degrees Celsius.
        """
        return None

    @property
    def smart_table(self):
        """
        Returns the SMART table for the device, if available.
        """
        return {}

    def get_filters(self) -> List[str]:
        """
        Returns a list of filters that should be used when looking up this
        device in the device database.
        """
        return []

    @abc.abstractmethod
    def __enter__(self):
        pass

    @abc.abstractmethod
    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


def get_device(path: Union[Path, str]) -> Device:
    """
    Returns a Device instance for the given path. This is a convenience
    function for the Device constructor which tries to automatically
    determine the correct subclass to use.

    .. code-block:: python

        with get_device("/dev/sda") as device:
            print(device.model, device.serial)

    :param path: The filesystem path to the device (such as /dev/sda).
    """
    # Remember, we always need to use delayed imports here as these imports
    # depend on the platform and may import things that aren't available on
    # all platforms.
    system = platform.system()
    if system == "Windows":
        from smartie.scsi.windows import WindowsSCSIDevice

        return WindowsSCSIDevice(path)
    elif system == "Linux":
        from smartie.nvme.linux import LinuxNVMeDevice
        from smartie.scsi.linux import LinuxSCSIDevice

        if "nvme" in str(path):
            return LinuxNVMeDevice(path)
        return LinuxSCSIDevice(path)
    else:
        raise NotImplementedError("Device not implemented for this platform.")


def get_all_devices() -> Iterable[Device]:
    """
    Yields all the devices detected on the host.

    .. code-block:: python

        for device in get_all_devices():
            with device:
                print(device.model, device.serial)
    """
    system = platform.system()
    if system == "Linux":
        # There's gotta be a better way of doing this, but I haven't found
        # it yet. We could ask udisk via dbus, or look at /dev/disks/*,
        # but the below gimmick works on the latest beta kernels and the
        # oldest linux kernels I could get my hands on.
        p = Path("/sys/block")
        if not p.exists or not p.is_dir():
            return

        for child in p.iterdir():
            # This whitelist is guaranteed to be the cause of a headache or
            # two, but I can't currently find a portable way of getting the
            # device "type". We don't want to get encrypted volumes or memory
            # disks, for example.
            if not child.name.startswith(("sd", "nvme")):
                continue

            # In the new kernel, nvme device may has a character device like
            # nvme0c0n1 (/sys/block/nvme0c0n1), but it is not in the directo
            # -ry /dev/.
            device_path = Path("/dev") / child.name
            if os.path.exists(device_path):
                yield get_device(device_path)
    elif system == "Windows":
        k32 = ctypes.WinDLL("kernel32", use_last_error=True)

        devices = ctypes.create_unicode_buffer(65536)
        # QueryDosDevice will return a list of NULL-terminated strings as a
        # binary blob. Each string is the name of a device (usually hundreds
        # on the typical desktop) that we may or may not care about.
        # The function returns the number of bytes it actually wrote to
        # `devices`.
        bytes_written = k32.QueryDosDeviceW(
            None, devices, ctypes.sizeof(devices)
        )
        if bytes_written == 0:
            raise RuntimeError("")

        i = 0
        while i < bytes_written:
            # Grab all the characters in the path until we get to the NULL
            # (0x00) byte.
            device_path = "".join(
                itertools.takewhile(lambda c: c != "\x00", devices[i:])  # noqa
            )
            i += len(device_path) + 1
            # Ignore every device that doesn't look like PhysicalDrive0, CdRom0,
            # PhysicalDrive1, etc...
            if device_path.startswith(("PhysicalDrive", "CdRom")):
                # PathLib cannot be used here, there's an option CPython ticket
                # for errors resolving device paths. This is unfortunate,
                # it forced us to revert to a string as the base path type.
                yield get_device(f"\\\\.\\{device_path}")
    elif system == "Darwin":
        from smartie.platforms.osx import iokit, cf, kCFBooleanTrue

        io_iterator = ctypes.c_void_p()

        query = iokit.IOServiceMatching(b"IOBlockStorageDevice")
        cf.CFDictionaryAddValue(
            query,
            cf.CFStringCreateWithCString(None, b"SMART Capable", 0),
            kCFBooleanTrue,
        )

        result = iokit.IOServiceGetMatchingServices(
            0,
            query,
            ctypes.byref(io_iterator),  # kIOMasterPortDefault
        )

        if result != 0:
            raise OSError(ctypes.get_errno())

        while iokit.IOIteratorIsValid(io_iterator):
            io_device = iokit.IOIteratorNext(io_iterator)
            if not io_device:
                break

            name = ctypes.create_string_buffer(512)

            iokit.IORegistryEntryGetPath(io_device, b"IOService", name)

            yield get_device(name.value)
    else:
        raise NotImplementedError("platform not supported")



import sys
import json
import ctypes

import click
from rich import box
from rich.console import Console, Group, group
from rich.table import Table

from smartie.database import DRIVE_DATABASE, get_matching_drive_entries
from smartie.device import get_all_devices, get_device
from smartie.nvme import NVMeDevice
from smartie.scsi import SCSIDevice
from smartie.structures import c_uint128, embed_bytes
from smartie.util import grouper_it


@group()
def print_structure(structure: ctypes.Structure, *, indent=0):
    """
    Pretty prints a ctypes.Structure.
    """
    offset = 0

    t = Table(show_lines=True)
    t.add_column("Offset", style="white italic", justify="left")
    t.add_column("Name", style="magenta")
    t.add_column("Value")

    for field in structure._fields_:  # noqa
        if len(field) == 3:
            # If the field has a 3rd part, it's a bitfield, with the 3rd part
            # being the bit count.
            name, type_, bitcount = field
        else:
            name, type_ = field
            bitcount = ctypes.sizeof(type_) * 8

        value = getattr(structure, name)

        if isinstance(value, ctypes.Array):
            array_table = Table(show_header=False)
            array_table.add_column("Hex", no_wrap=True, style="green")
            array_table.add_column("ASCII", no_wrap=True, style="white")

            for chunk in grouper_it(20, bytearray(value)):
                chunk = list(chunk)

                array_table.add_row(
                    " ".join(f"{byte:02X}" for byte in chunk),
                    "".join(
                        chr(byte) if 32 <= byte <= 126 else "."
                        for byte in chunk
                    ),
                )

            t.add_row(
                f"[{offset:03}:{offset + bitcount:03}]", name, array_table
            )
        elif isinstance(value, c_uint128):
            t.add_row(
                f"[{offset:03}:{offset + bitcount:03}]",
                name,
                f"0x{int(value):03X}",
            )
        elif isinstance(value, ctypes.Structure):
            t.add_row(
                f"[{offset:03}:{offset + bitcount:03}]",
                name,
                Group(print_structure(value, indent=indent + 2)),
            )
        elif value is None:
            t.add_row(
                f"[{offset:03}:{offset + bitcount:03}]",
                name,
                "None",
            )
        else:
            t.add_row(
                f"[{offset:03}:{offset + bitcount:03}]",
                name,
                f"0x{int(value):03X}",
            )

        offset += bitcount

    yield t


@click.group()
def cli():
    """
    Command line interface for SMARTie.
    """


@cli.command("enumerate")
def enumerate_command():
    """
    Enumerate all available devices, displaying basic information.
    """
    table = Table(box=box.SIMPLE)
    table.add_column("Path", style="magenta")
    table.add_column("Model", style="green")
    table.add_column("Serial", style="blue")
    table.add_column("Temperature")

    for device in get_all_devices():
        with device:
            table.add_row(
                device.path,
                device.model,
                device.serial,
                f"{device.temperature}",
            )

    console = Console()
    console.print(table)


@cli.command("details")
@click.argument("path")
def details_command(path: str):
    """
    Show detailed information for a specific device.
    """

    def blocks_to_gb(blocks: int) -> float:
        return blocks * 1000 * 512 / 1000 / 1000 / 1000

    details_table = Table(show_header=False)
    details_table.add_column("Key", style="magenta")
    details_table.add_column("Value", style="green")

    with get_device(path) as device:
        details_table.add_row("Model Number", device.model)
        details_table.add_row("Serial Number", device.serial)
        details_table.add_row("Temperature", f"{device.temperature}C")

        smart_table = Table(title="SMART Attributes", title_style="magenta")
        if isinstance(device, SCSIDevice):
            smart_table.add_column("ID", style="white")
            smart_table.add_column("Name", style="magenta")
            smart_table.add_column("Current", style="green", justify="right")
            smart_table.add_column("Worst", style="blue", justify="right")
            smart_table.add_column("Threshold", style="yellow", justify="right")
            smart_table.add_column("Unit", style="italic white")

            for entry in device.smart_table.values():
                smart_table.add_row(
                    str(entry.id),
                    entry.name,
                    str(entry.current_value),
                    str(entry.worst_value),
                    str(entry.threshold),
                    entry.unit.name,
                )
        elif isinstance(device, NVMeDevice):
            smart_table.add_column("Name", style="magenta")
            smart_table.add_column("Value", style="green", justify="right")

            smart, _ = device.smart()

            # We only show a selection of attributes, as the full list is
            # not terribly useful.
            smart_table.add_row(
                "Critical Warning",
                print_structure(smart.critical_warning),
            )
            smart_table.add_row(
                "Temperature", f"{smart.temperature - 273.15:.2f}C"
            )
            smart_table.add_row("Available Spare", f"{smart.available_spare}%")
            smart_table.add_row(
                "Available Spare Threshold",
                f"{smart.available_spare_threshold}%",
            )
            smart_table.add_row("Percentage Used", f"{smart.percent_used}%")
            smart_table.add_row(
                "Data Units Read",
                f"{blocks_to_gb(int(smart.data_units_read)):.2f}GB",
            )
            smart_table.add_row(
                "Data Units Written",
                f"{blocks_to_gb(int(smart.data_units_written)):.2f}GB",
            )
            smart_table.add_row(
                "Host Read Commands", f"{smart.host_read_commands}"
            )
            smart_table.add_row(
                "Host Write Commands", f"{smart.host_write_commands}"
            )
            smart_table.add_row(
                "Controller Busy Time", f"{smart.controller_busy_time}"
            )
            smart_table.add_row("Power Cycles", f"{smart.power_cycles}")
            smart_table.add_row("Power On Hours", f"{smart.power_on_hours}")
            smart_table.add_row("Unsafe Shutdowns", f"{smart.unsafe_shutdowns}")
            smart_table.add_row(
                "Media and Data Integrity Errors", f"{smart.media_errors}"
            )
            smart_table.add_row(
                "Error Information Log Entries", f"{smart.num_err_log_entries}"
            )

            temp_table = Table(expand=True)
            temp_table.add_column("Sensor", style="magenta", justify="center")
            temp_table.add_column(
                "Temperature", style="magenta", justify="center"
            )

            for i, sensor in enumerate(smart.temperature_sensors):
                if sensor != 0x00:
                    temp_table.add_row(str(i), f"{int(sensor - 273.15)}C")

            smart_table.add_row("Temperature Sensors", temp_table)

        details_table.add_row("", smart_table)

    console = Console()
    console.print(details_table)


@cli.command("dump")
@click.argument("path")
@click.argument(
    "command", type=click.Choice(["inquiry", "identify", "smart", "thresholds"])
)
@click.option(
    "--display",
    default="pretty",
    type=click.Choice(["pretty", "raw", "bytearray"]),
)
def dump_command(path: str, command: str, display: str = "pretty"):
    """
    Dump raw responses from an NVMe or ATA device.

    This command can pretty-print rich structures, write raw bytes to stdout,
    or write a bytearray ready for embedding in Python to stdout. Control the
    output with the --display option.
    """
    console = Console()

    with get_device(path) as device:
        if isinstance(device, SCSIDevice):
            result = {
                "inquiry": device.inquiry,
                "identify": device.identify,
                "smart": device.smart,
                "thresholds": device.smart_thresholds,
            }.get(command)
            if result is None:
                console.print("Command unknown or unsupported by this device.")
                return

            structure = result()[0]
        elif isinstance(device, NVMeDevice):
            result = {"identify": device.identify, "smart": device.smart}.get(
                command
            )
            if result is None:
                console.print("Command unknown or unsupported by this device.")
                return

            structure = result()[0]
        else:
            raise NotImplementedError("Unknown device type.")

        if display == "pretty":
            console.print(print_structure(structure))
        elif display == "raw":
            sys.stdout.buffer.write(bytes(structure))  # noqa
        elif display == "bytearray":
            print(embed_bytes(bytearray(structure)))  # noqa


@cli.group("db")
def db_group():
    """
    Commands for querying the disk database.
    """


@db_group.command("export")
def db_export_command():
    """
    Export the disk database as JSON to stdout.

    This is useful for embedding the database in other applications.
    """
    print(
        json.dumps(
            [
                {
                    "name": entry.name,
                    "filters": [
                        f if isinstance(f, str) else f.pattern
                        for f in entry.filters
                    ],
                    "smart_attributes": [
                        {
                            "id": attr.id,
                            "name": attr.name,
                            "unit": attr.unit.name,
                        }
                        for attr in entry.smart_attributes.values()
                    ],
                }
                for entry in DRIVE_DATABASE
            ],
            indent=4,
            sort_keys=True,
        )
    )


@db_group.command("matches")
@click.argument("path")
def db_matches_command(path: str):
    """
    Show all matching entries in the disk database for a specific device.
    """
    console = Console()

    t = Table(box=box.SIMPLE)
    t.add_column("Name", style="magenta", vertical="middle")
    t.add_column("Filters", style="green")

    with get_device(path) as device:
        matches = get_matching_drive_entries(device.get_filters())
        for match in matches:
            filter_table = Table(show_header=False)
            filter_table.add_column("Type", style="white")
            filter_table.add_column("Filter", style="green")

            for f in match.filters:
                filter_table.add_row(
                    f.__class__.__name__,
                    f if isinstance(f, str) else f.pattern,  # noqa
                )

            t.add_row(match.name, filter_table)

    console.print(t)


@cli.group("api")
def api_group():
    """
    Commands for interacting with SMARTie from other applications, such as
    shell scripts.
    """


@api_group.command("list")
def api_list_command():
    """
    List all devices in the system.
    """
    results = []
    for device in get_all_devices():
        with device:
            results.append(
                {
                    "path": device.path,
                    "model": device.model,
                    "serial": device.serial,
                    "temperature": device.temperature,
                }
            )

    print(json.dumps(results, indent=4, sort_keys=True))


@api_group.command("get")
@click.argument("path")
def api_get_command(path: str):
    """
    Get detailed information about a specific device.
    """
    result = {}
    with get_device(path) as device:
        result["path"] = device.path
        result["model"] = device.model
        result["serial"] = device.serial
        result["temperature"] = device.temperature

        if isinstance(device, SCSIDevice):
            result["smart"] = {}
            for entry in device.smart_table.values():
                result["smart"][entry.id] = {
                    "id": entry.id,
                    "current": entry.current_value,
                    "worst": entry.worst_value,
                    "threshold": entry.threshold,
                    "unit": entry.unit.name,
                    "flags": entry.flags,
                }
        elif isinstance(device, NVMeDevice):
            result["smart"] = device.smart_table

    print(json.dumps(result, indent=4, sort_keys=True))



import ctypes
from typing import Any, Dict

from smartie.util import grouper_it


class c_uint128(ctypes.Structure):  # noqa
    """
    Utility shim to make it easier to work with 128-bit integers.
    """

    _pack_ = 1
    _fields_ = [("low", ctypes.c_uint64), ("high", ctypes.c_uint64)]

    def __init__(self, value: int = 0):
        super().__init__()
        self.low = value & 0xFFFFFFFFFFFFFFFF
        self.high = (value >> 64) & 0xFFFFFFFFFFFFFFFF

    @property
    def value(self) -> int:
        return (self.high << 64) | self.low

    def __int__(self):
        return self.value

    def __format__(self, format_spec):
        return self.value.__format__(format_spec)


def swap_bytes(src):
    # Weirdly, all the strings in the IDENTIFY response are byte swapped.
    src = bytearray(src)

    for i in range(0, len(src) - 1, 2):
        src[i] ^= src[i + 1]
        src[i + 1] ^= src[i]
        src[i] ^= src[i + 1]

    return src


def swap_int(c: int, n: int) -> int:
    return int.from_bytes(
        n.to_bytes(c, byteorder="little"), byteorder="big", signed=False
    )


def embed_bytes(data: bytes, *, indent=0, char="    ", max_width=80) -> str:
    """
    Pretty-prints `data` in such a way that it can be embedded cleanly in
    a Python file.

    This exists to embed SCSI commands and responses into tests.

    :param data: The binary data to be formatted.
    :param indent: The number of characters to indent each line.
    :param char: The character to use for indentation.
    :param max_width: The maximum length of each line.
    :return: The formatted result.
    """
    prefix = char * indent
    line_length = max_width - len(prefix)

    lines = "\n".join(
        "{prefix}{line}".format(
            prefix=char * (indent + 1),
            line=", ".join(f"0x{byte:02X}" for byte in row),
        )
        for row in grouper_it(line_length // 6, data)
    )
    return f"{prefix}bytearray([\n{lines}\n{prefix}])"


def pprint_structure(s: ctypes.Structure):
    """
    Debugging utility method to pretty-print a `ctypes.Structure`.
    """
    offset = 0

    for field in s._fields_:  # noqa
        if len(field) == 3:
            # Found a bitfield.
            name, type_, bitcount = field
        else:
            name, type_ = field
            bitcount = ctypes.sizeof(type_) * 8

        value = getattr(s, name)
        if isinstance(value, ctypes.Array):
            print(
                f"{name}[{offset}:{offset + bitcount}] = {bytes(value)[:15]!r}"
                f" ({len(value)} bytes)"
            )
            print(embed_bytes(bytes(value)))
        else:
            print(f"{name}[{offset}:{offset + bitcount}] = {value!r}")
        offset += bitcount


def structure_to_dict(s: ctypes.Structure) -> Dict[str, Any]:
    """
    Converts a `ctypes.Structure` into a dictionary.
    """
    result = {}

    for field in s._fields_:  # noqa
        name = field[0]
        value = getattr(s, name)

        if isinstance(value, ctypes.Array):
            result[name] = [v for v in value]
        elif isinstance(value, c_uint128):
            result[name] = int(value)
        elif isinstance(value, ctypes.Structure):
            result[name] = structure_to_dict(value)
        else:
            result[name] = value

    return result



from smartie.nvme.constants import NVMe_STATUS_FIELD


class NVMeStatusFieldError(Exception):
    """
    Raised when an NVMe command returns an error status field.
    """

    def __init__(
        self,
        status_code: int,
        status_code_type: int,
        *,
        status_field=None,
    ):
        super().__init__()
        #: The status code.
        self.status_code = status_code
        #: The status code type.
        self.status_code_type = status_code_type
        #: The status field structure, if available.
        self.status_filed = status_field

    @property
    def error_msg(self) -> str:
        return NVMe_STATUS_FIELD.get(
            (self.status_code_type, self.status_code),
            "Unknown Error",
        )

    def __repr__(self):
        return (
            f"<{self.__class__.__name__}("
            f"status_code_type=0x{self.status_code_type:02x},"
            f" status_code=0x{self.status_code:02x},"
            f" err={self.error_msg!r}"
            f")>"
        )



import ctypes
from functools import cache


@cache
def get_kernel32():
    # Opens the Kernel32.dll, which can be quite a slow process, and
    # saves it for future use.
    return ctypes.WinDLL("kernel32", use_last_error=True)



"""
Ancillary support functions for OS X.
"""

import ctypes

iokit = ctypes.CDLL(
    "/System/Library/Frameworks/IOKit.framework/IOKit", use_errno=True
)

cf = ctypes.CDLL(
    "/System/Library/Frameworks/CoreFoundation.framework/CoreFoundation",
    use_errno=True,
)

kCFBooleanTrue = ctypes.c_void_p.in_dll(cf, "kCFBooleanTrue")

iokit.IOServiceMatching.restype = ctypes.c_void_p

iokit.IOServiceGetMatchingServices.restype = ctypes.c_int
iokit.IOServiceGetMatchingServices.argtypes = [
    ctypes.c_void_p,
    ctypes.c_void_p,
    ctypes.c_void_p,
]

iokit.IORegistryEntryCreateCFProperty.restype = ctypes.c_void_p
iokit.IORegistryEntryCreateCFProperty.argtypes = [
    ctypes.c_void_p,
    ctypes.c_void_p,
    ctypes.c_void_p,
    ctypes.c_uint32,
]

cf.CFStringCreateWithCString.restype = ctypes.c_void_p
cf.CFStringCreateWithCString.argtypes = [
    ctypes.c_void_p,
    ctypes.c_char_p,
    ctypes.c_uint32,
]

cf.CFDictionaryAddValue.restype = None
cf.CFDictionaryAddValue.argtypes = [
    ctypes.c_void_p,
    ctypes.c_void_p,
    ctypes.c_void_p,
]



"""
Ancillary support functions for Linux.
"""

import ctypes
from functools import cache


@cache
def get_libc():
    # Opens the libc.so, which can be quite a slow process, and
    # saves it for future use.
    return ctypes.CDLL("libc.so.6", use_errno=True)



__all__ = ("SCSIDevice", "SCSIResponse")

import abc
import ctypes
from dataclasses import replace, dataclass
from typing import Dict, List, Optional, Tuple, Union, Any

import smartie.structures
from smartie.database import SMARTAttribute, get_drive_entry
from smartie.device import Device
from smartie.scsi.errors import SenseError
from smartie.scsi.structures import (
    ATACommands,
    ATAPICommands,
    ATAProtocol,
    ATASmartFeature,
    Command16,
    CommandFlags,
    DescriptorFormatSense,
    DeviceType,
    Direction,
    FixedFormatSense,
    IdentifyResponse,
    InquiryCommand,
    InquiryResponse,
    OperationCode,
    SmartDataResponse,
    SmartThresholdResponse,
    Command12,
)
from smartie.structures import swap_bytes


@dataclass
class SCSIResponse:
    """
    Common response object for SCSI commands.

    This object attempts to encapsulate the response from an SCSI command in a
    platform-agnostic way. It contains the sense data, the command that was
    issued, and whether the command succeeded or not.

    For additional platform-specific information, the `platform_header`
    attribute contains the platform-specific header that was used to issue the
    command.
    """

    #: Whether the command succeeded. If None, the status is unknown.
    succeeded: Optional[bool]
    #: The sense data returned by the device.
    sense: Optional[Union[FixedFormatSense, DescriptorFormatSense]]
    #: The command issued to the device.
    command: Union[Command16, Command12]
    #: The actual number of bytes transferred.
    bytes_transferred: Optional[int]

    #: The platform-specific header that was used to issue the command.
    #: For example this may be an :class:`SCSIPassThroughDirectWithBuffer` on
    #: Windows.
    platform_header: Any

    def __bool__(self):
        return self.succeeded


class SCSIDevice(Device, abc.ABC):
    @classmethod
    def parse_sense(
        cls, sense_blob
    ) -> Optional[Union[FixedFormatSense, DescriptorFormatSense]]:
        """
        Parses the sense response from an SCSI command, raising a
        :class:`smartie.errors.SenseError` if an error occurred.

        Will return either a :class:`structures.FixedFormatSense` or a
        :class:`structures.DescriptorFormatSense` depending on the error code.

        :param sense_blob: A bytearray (or similar) object containing the
                           unparsed sense response.
        """
        error_code = sense_blob[0] & 0x7F
        if error_code == 0x00:
            return None
        elif error_code in (0x70, 0x71):
            sense = FixedFormatSense.from_buffer_copy(sense_blob)
            if sense.sense_key not in (0x00, 0x01, 0x0F):
                raise SenseError(sense.sense_key, sense=sense)
            return sense
        elif error_code in (0x72, 0x73):
            sense = DescriptorFormatSense.from_buffer_copy(sense_blob)
            if sense.sense_key not in (0x00, 0x01, 0x0F):
                raise SenseError(sense.sense_key, sense=sense)
            return sense
        else:
            raise SenseError(0, sense=sense_blob)

    def issue_command(
        self,
        direction: Direction,
        command: ctypes.Structure,
        data: Union[ctypes.Array, ctypes.Structure, None],
        *,
        timeout: int = 3000,
    ) -> SCSIResponse:
        """
        Issues an SCSI passthrough command to the disk.

        :param direction: Direction for this command.
        :param command: Command to be sent to the device.
        :param data: Command data to be sent/received to/from the device.
        :param timeout: Timeout in milliseconds. Setting this to MAX_INT
                        results in no timeout.
        """
        raise NotImplementedError()

    def inquiry(self) -> Tuple[InquiryResponse, SCSIResponse]:
        """
        Issues a standard SCSI INQUIRY command.
        """
        inquiry = InquiryResponse()

        inquiry_command = InquiryCommand(
            operation_code=OperationCode.INQUIRY, allocation_length=96
        )

        response = self.issue_command(Direction.FROM, inquiry_command, inquiry)

        return inquiry, response

    def identify(
        self, try_atapi_on_failure=True
    ) -> Tuple[IdentifyResponse, SCSIResponse]:
        """
        Issues a standard ATA IDENTIFY command.

        :param try_atapi_on_failure: If True, will try an ATAPI IDENTIFY command
                                        if the ATA IDENTIFY command fails.
        """
        identity = ctypes.create_string_buffer(b"\x00", 512)

        command16 = Command16(
            operation_code=OperationCode.COMMAND_16,
            protocol=ATAProtocol.PIO_DATA_IN << 1,
            flags=CommandFlags(
                t_length=CommandFlags.Length.IN_SECTOR_COUNT,
                byt_blok=True,
                t_dir=True,
                ck_cond=True,
            ),
            command=ATACommands.IDENTIFY,
        )

        try:
            response = self.issue_command(Direction.FROM, command16, identity)
        except SenseError:
            # If an error occurred, we try to see if this is really an ATAPI
            # device, such as a CD-ROM. We can handle the response exactly
            # the same, it's just a different command.
            if not try_atapi_on_failure:
                raise

            command16.command = ATAPICommands.IDENTIFY
            response = self.issue_command(Direction.FROM, command16, identity)

        return IdentifyResponse.from_buffer(identity), response

    @property
    def model(self) -> Optional[str]:
        """
        Returns the model name of the device.
        """
        identity, response = self.identify()
        v = swap_bytes(identity.model_number).strip(b" \x00").decode()
        # If we didn't get anything at all back from an ATA IDENTIFY, try an
        # old fashion SCSI INQUIRY.
        if not v:
            inquiry, response = self.inquiry()
            v = (
                bytearray(inquiry.product_identification)
                .strip(b" \x00")
                .decode()
            )
        return v

    @property
    def serial(self) -> Optional[str]:
        """
        Returns the serial number of the device.
        """
        identity, response = self.identify()
        v = swap_bytes(identity.serial_number).strip(b" \x00").decode()
        # If we didn't get anything at all back from an ATA IDENTIFY, try an
        # old fashion SCSI INQUIRY.
        if not v:
            inquiry, response = self.inquiry()
            v = bytearray(inquiry.vendor_specific_1).strip(b" \x00").decode()
        return v

    @property
    def temperature(self) -> Optional[int]:
        """
        Returns the temperature of the device in degrees Celsius.
        """
        smart_table = self.smart_table

        temp = smart_table.get(0xBE)
        if temp is not None:
            return temp.p_value

        temp = smart_table.get(0xC2)
        if temp is not None:
            return temp.p_value

    @property
    def device_type(self):
        """
        Get the device's type, if available.
        """
        inquiry, sense = self.inquiry()
        return DeviceType(inquiry.peripheral_device_type)

    def smart_thresholds(self) -> Tuple[SmartThresholdResponse, SCSIResponse]:
        """
        Issues an ATA SMART READ_THRESHOLDS command.
        """
        thresholds = SmartThresholdResponse()

        command16 = Command16(
            operation_code=OperationCode.COMMAND_16,
            protocol=ATAProtocol.PIO_DATA_IN << 1,
            command=ATACommands.SMART,
            flags=CommandFlags(
                t_length=CommandFlags.Length.IN_SECTOR_COUNT,
                byt_blok=True,
                t_dir=True,
                ck_cond=True,
            ),
            features=smartie.structures.swap_int(
                2, ATASmartFeature.SMART_READ_THRESHOLDS
            ),
        ).set_lba(0xC24F00)

        response = self.issue_command(Direction.FROM, command16, thresholds)
        return thresholds, response

    def smart(self) -> Tuple[SmartDataResponse, SCSIResponse]:
        """
        Issues an ATA SMART READ_DATA command.
        """
        smart = SmartDataResponse()

        command16 = Command16(
            operation_code=OperationCode.COMMAND_16,
            protocol=ATAProtocol.PIO_DATA_IN << 1,
            command=ATACommands.SMART,
            flags=CommandFlags(
                t_length=CommandFlags.Length.IN_SECTOR_COUNT,
                byt_blok=True,
                t_dir=True,
                ck_cond=True,
            ),
            features=smartie.structures.swap_int(
                2, ATASmartFeature.SMART_READ_DATA
            ),
        ).set_lba(0xC24F00)

        response = self.issue_command(Direction.FROM, command16, smart)
        return smart, response

    def get_filters(self) -> List[str]:
        return ["type:ata", f"model:{self.model}"]

    @property
    def smart_table(self) -> Dict[int, SMARTAttribute]:
        """
        Returns a parsed and processed dictionary of SMART attributes.
        """
        drive_entry = get_drive_entry(self.get_filters())

        thresholds, _ = self.smart_thresholds()
        p_thresholds = {}
        for entry in thresholds.entries:
            if entry.attribute_id == 0x00:
                break
            p_thresholds[entry.attribute_id] = entry.value

        smart, _ = self.smart()
        result = {}
        for entry in smart.attributes:
            if entry.id == 0x00:
                break

            result[entry.id] = replace(
                drive_entry.smart_attributes.get(
                    entry.id,
                    SMARTAttribute(
                        "UNKNOWN",
                        id=entry.id,
                        flags=entry.flags,
                        current_value=entry.current,
                        worst_value=entry.worst,
                        threshold=p_thresholds.get(entry.id),
                    ),
                ),
                flags=entry.flags,
                current_value=entry.current,
                worst_value=entry.worst,
                threshold=p_thresholds.get(entry.id),
            )

        return result



import itertools


def grouper_it(n, iterable):
    it = iter(iterable)
    while True:
        chunk_it = itertools.islice(it, n)
        try:
            first_el = next(chunk_it)
        except StopIteration:
            return
        yield itertools.chain((first_el,), chunk_it)



import ctypes
from typing import Union

from smartie.scsi import SCSIDevice, SCSIResponse
from smartie.platforms.win32 import get_kernel32
from smartie.scsi.structures import (
    Direction,
    IOCTL_SCSI_PASS_THROUGH_DIRECT,
    SCSIPassThroughDirect,
    SCSIPassThroughDirectWithBuffer,
)


class WindowsSCSIDevice(SCSIDevice):
    def __enter__(self):
        if self.fd is not None:
            raise IOError("Device is already open.")

        # We can't use the normal approach to opening a file on Windows, as
        # various Python APIs can't handle a device opened without specific
        # flags, see (https://bugs.python.org/issue37074)
        self.fd = get_kernel32().CreateFileW(
            self.path,
            0x80000000 | 0x40000000,  # GENERIC_READ | GENERIC_WRITE
            0x00000001,  # FILE_SHARE_READ
            None,
            0x00000003,  # OPEN_EXISTING
            0x00000080,  # FILE_ATTRIBUTE_NORMAL,
            None,
        )

        if self.fd == -1:
            raise ctypes.WinError(ctypes.get_last_error())

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.fd is not None:
            get_kernel32().CloseHandle(self.fd)
            self.fd = None
        return False

    def issue_command(
        self,
        direction: Direction,
        command: ctypes.Structure,
        data: Union[ctypes.Array, ctypes.Structure, None],
        *,
        timeout: int = 3000,
    ) -> SCSIResponse:
        # On Windows, the command block is always 16 bytes, but we may be
        # sending a smaller command. We use a temporary mutable bytearray for
        # this.
        cdb = (ctypes.c_ubyte * 16).from_buffer_copy(
            bytearray(command).ljust(16, b"\x00")  # noqa
        )

        if data is None:
            data = ctypes.create_string_buffer(0)

        header_with_buffer = SCSIPassThroughDirectWithBuffer(
            sptd=SCSIPassThroughDirect(
                length=ctypes.sizeof(SCSIPassThroughDirect),
                data_in={
                    Direction.TO: 0,
                    Direction.FROM: 1,
                    Direction.NONE: 2,
                }.get(direction),
                data_transfer_length=ctypes.sizeof(data),
                data_buffer=ctypes.addressof(data),
                cdb_length=ctypes.sizeof(command),
                cdb=cdb,
                timeout_value=max(timeout // 1000, 1),
                sense_info_length=SCSIPassThroughDirectWithBuffer.sense.size,
                sense_info_offset=(
                    SCSIPassThroughDirectWithBuffer.sense.offset
                ),
            )
        )

        result = get_kernel32().DeviceIoControl(
            self.fd,
            IOCTL_SCSI_PASS_THROUGH_DIRECT,
            ctypes.pointer(header_with_buffer),
            ctypes.sizeof(header_with_buffer),
            ctypes.pointer(header_with_buffer),
            ctypes.sizeof(header_with_buffer),
            None,
            None,
        )

        self.parse_sense(bytearray(header_with_buffer.sense))

        if result == 0:
            raise ctypes.WinError(ctypes.get_last_error())

        return SCSIResponse(
            succeeded=(
                header_with_buffer.sptd.scsi_status == 0
                or header_with_buffer.sptd.scsi_status == 2
            ),
            sense=self.parse_sense(bytearray(header_with_buffer.sense)),
            platform_header=header_with_buffer,
            command=command,
            bytes_transferred=header_with_buffer.sptd.data_transfer_length,
        )
